package main.java;

import javafx.application.Application;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Insets;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Slider;
import javafx.scene.control.TextField;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.HBox;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import main.java.datamodel.Person;
import main.java.datamodel.Relationship;
import main.java.datamodel.RelationshipType;
import main.java.datamodel.Sex;
import main.java.datamodel.exceptions.WrongFilePathException;
import main.java.layout.LocalBlock;
import main.java.layout.GedComConversion;
import main.java.shape.*;
import main.java.shape.abs.DraggableShape;

import java.io.File;
import java.sql.Time;
import java.util.*;
import java.util.function.Consumer;

public class Main extends Application {

    private Stage stage;
    private CustomBorderPane customBorderPane;
    public static String path;
    private List<LocalBlock> blocks;
    private Slider slider = new Slider();
    public static LifeLine selectedLifeline;
    private TimeAxis timeAxis;
    private HBox hBox;
    private Scene scene;
    private int searchIndex = 0;

    /**
     * The start of the javaFX application.
     * @param primaryStage
     * @throws Exception
     */
    @Override
    public void start(Stage primaryStage) throws Exception{
        this.stage = primaryStage;
        primaryStage.setTitle("TimeNets");
        scene = new Scene(createPane(), 800, 600);
        primaryStage.setScene(scene);
        primaryStage.show();
        primaryStage.setMaximized(true);
    }

    /**
     * Creates the border pane and adds all needed functionality.
     * @return the created border pane.
     * @throws WrongFilePathException if a problem occurs with the file.
     */
    private Parent createPane() throws WrongFilePathException {

        customBorderPane = CustomBorderPane.instance();

        hBox = new HBox();
        hBox.setSpacing(10);
        hBox.setPadding(new Insets(10));
        Button openButton = new Button("Open");
        openButton.setOnAction((e) -> {
            try {
                this.open();
            } catch (WrongFilePathException e1) {
                e1.printStackTrace();
            }
        });

        TextField searchTextField = new TextField();
        searchTextField.setPromptText("search...");
        searchTextField.setOnAction((e) -> {
            search(searchTextField.getText());
        });


        slider.setMin(0);
        slider.setMax(100);
        slider.setValue(0);
        slider.setShowTickLabels(true);
        slider.setShowTickMarks(true);
        slider.setMajorTickUnit(50);
        slider.setMinorTickCount(5);
        slider.setBlockIncrement(10);
        slider.valueProperty().addListener((observable, oldValue, newValue) -> {

            //textField.setText(Double.toString(newValue.intValue()));
//            CustomBorderPane.instance().getChildren().removeAll();
//            CustomBorderPane.removeObjects();

            if(path != null){
                drawOnPaneGedComConverted();
            }

        });

        hBox.getChildren().addAll(openButton, searchTextField,slider);
        customBorderPane.setTop(hBox);

        SelectionHandler selectionHandler = new SelectionHandler(customBorderPane);
        customBorderPane.addEventHandler(MouseEvent.MOUSE_PRESSED, selectionHandler.getMousePressedEventHandler());

        drawOnPane(null);

        return customBorderPane;
    }


    /**
     * handles the drawing on the pane if the gedcom is already converted
     * adjusts the timeline to the converted structure, by choosing the minimal and maximal dates of the geneaological structure
     */
    private void drawOnPaneGedComConverted() {

        customBorderPane.getChildren().removeIf(child -> !(child instanceof HBox));


        List<Person> persons = new ArrayList<>();
        for(LocalBlock localBlock : blocks){
            persons.addAll(localBlock);
        }

        persons.sort((o1, o2) -> o1.getDate_of_birth().before(o2.getDate_of_birth()) ? -1 : o1.getDate_of_birth().after(o2.getDate_of_birth()) ? 1 : 0);

        Person oldest = persons.get(0);
        Date oldestBirthDate = oldest.getDate_of_birth();
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(oldestBirthDate);
        int begin = calendar.get(Calendar.YEAR);



        Date now = new Date();
        calendar.setTime(now);
        int end = calendar.get(Calendar.YEAR);


        timeAxis = new TimeAxis(begin - 10, end + 10);
        timeAxis.setHeightProperty(scene.heightProperty());

        createLifeLines(blocks,timeAxis);

    }


    /**
     * Draws all the data on the pane which is contained in the file specified by the path.
     * adjusts the timeline to the converted structure, by choosing the minimal and maximal dates of the geneaological structure
     * @param path the path of the gedcom file that should be visualized.
     * @throws WrongFilePathException if a problem occurs with the file.
     */
    private void drawOnPane(String path) throws WrongFilePathException {

        if(path == null)
            return;
        LinkedList<LocalBlock> localBlocks = GedComConversion.getLocalBlocks(path);

        List<Person> persons = new ArrayList<>();
        for(LocalBlock localBlock : localBlocks){
            persons.addAll(localBlock);
        }

        persons.sort((o1, o2) -> o1.getDate_of_birth().before(o2.getDate_of_birth()) ? -1 : o1.getDate_of_birth().after(o2.getDate_of_birth()) ? 1 : 0);

        Person oldest = persons.get(0);
        Date oldestBirthDate = oldest.getDate_of_birth();
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(oldestBirthDate);
        int begin = calendar.get(Calendar.YEAR);

        Date now = new Date();
        calendar.setTime(now);
        int end = calendar.get(Calendar.YEAR);


        timeAxis = new TimeAxis(begin - 10, end + 10);
        timeAxis.setHeightProperty(scene.heightProperty());

        createLifeLines(localBlocks,timeAxis);
    }


    /**
     * The main method.
     * @param args the string arguments.
     */
    public static void main(String[] args) {
        launch(args);
    }

    /**
     * Method that is called if the open button is pressed.
     * @throws WrongFilePathException if a problem occurs with the selected file.
     */
    private void open() throws WrongFilePathException {
        FileChooser fileChooser = new FileChooser();
        fileChooser.setTitle("Open Resource File");
        fileChooser.getExtensionFilters().addAll(
                new FileChooser.ExtensionFilter("GedCom Files", "*.ged"));
        File selectedFile = fileChooser.showOpenDialog(stage);
        if (selectedFile != null) {
            path = selectedFile.getAbsolutePath();
            blocks = null;
            selectedLifeline = null;
            CustomBorderPane.removeObjects();
            customBorderPane.getChildren().removeIf(child -> !(child instanceof HBox));
            drawOnPane(selectedFile.getAbsolutePath());
        }
    }

    /**
     * method that is called if the search text field is used.
     * @param searchString The string which contains the name of the wanted person.
     */
    private void search(String searchString){

        if(CustomBorderPane.DRAGGABLE_SHAPES.isEmpty())
            return;

        for(int i = searchIndex+1; i != searchIndex; i++){

            if(i >= CustomBorderPane.DRAGGABLE_SHAPES.size()){
                i = 0;
            }
            DraggableShape draggableShape = CustomBorderPane.DRAGGABLE_SHAPES.get(i);
            if(draggableShape.getClass() == LifeLine.class){
                LifeLine lifeLine = (LifeLine) draggableShape;
                lifeLine.notifySelection(false);

                String name = lifeLine.getName().toLowerCase();
                String search = searchString.toLowerCase();


                if(name.contains(search)){
                    selectedLifeline = lifeLine;
                    lifeLine.notifySelection(true);
                    double x = lifeLine.getStartX();
                    double y = lifeLine.getStartY();
                    CustomBorderPane.DRAGGABLE_SHAPES.forEach(draggableShape1 -> {
                        draggableShape1.dragX(CustomBorderPane.instance().startDragX - x + 100);
                        draggableShape1.dragY(CustomBorderPane.instance().startDragY - y + 100);
                    });
                    searchIndex = i+1;
                    decreaseDOIRoot(lifeLine.getPerson());
                    break;
                }
            }
        }
    }

    /**
     * Decreases the Degree of Interest starting from a specific person
     * normally this person will be a person of interest. Probably selcted by clicking on the lifeline or by searching  with the searchbar
     * @param person the root/ start person
     */
    public static void decreaseDOIRoot(Person person){
        //starting from Person person the DOI values will be decreased
        // persons which are not related to the person at all will get the DOI value 0
        person.setDoiVisited(true);

        int spousalCounter = 10;// the first spouse has a higher DOI than later ones
        int childDecreaseValue = 10;
        //todo: Lukas please check if this is correct. I hope so...
        for (Relationship r : person.getRelationships()) {
            if(r.getType() == RelationshipType.MARRIAGE || r.getType() == RelationshipType.DIVORCE){
                decreaseDOI(r.getPerson1(), spousalCounter, 1);
                decreaseDOI(r.getPerson2(), spousalCounter, 1);
            }else{
                decreaseDOI(r.getPerson1(), childDecreaseValue, 1);
                decreaseDOI(r.getPerson2(), childDecreaseValue, 1);
                decreaseDOI(r.getPerson3(), childDecreaseValue, 1);
            }
            spousalCounter+=10;
        }

    }

    /**
     * Traverses the relationships and decreases the Degree of Interest values. Depending on the depth the DoI will be decreased more
     * @param person the person who should have a decreased DoI
     * @param value how much should the DOI be decreased ( multiplied with the counter)
     * @param counter counts the depth of the DFS
     */
    private static void decreaseDOI(Person person, int value, int counter){

        if(person == null || person.isDoiVisited()) //if the person was already visited while traversing and calculating the DOI
            return;

        person.setCurrentDOI(person.getCurrentDOI() - value*counter);
        person.setDoiVisited(true);

        for (Relationship r : person.getRelationships()) {
            decreaseDOI(r.getPerson1(), value, ++counter);
            decreaseDOI(r.getPerson2(), value, ++counter);
            decreaseDOI(r.getPerson3(), value, ++counter);
        }
    }


    /**
     * Constructs the life lines of all persons in the local blocks and places them in the layout.
     * @param localBlocks a list of all local blocks which contain all persons that should be rendered.
     * @param timeAxis the time axis of the layout.
     */
    private void createLifeLines(List<LocalBlock> localBlocks, TimeAxis timeAxis){

        this.blocks = localBlocks;

        double currentHeight = 30;
        double yDiff = 50;
        Map<Person, LifeLine> lifeLinesOfPersons = new HashMap<>();
        Map<Person, Double> initYPosition = new HashMap<>();
        Map<Person, Double> currentYPosition = new HashMap<>();
        Map<Person, Integer> spouseCountAbove = new HashMap<>();
        Map<Person, Integer> spouseCountBelow = new HashMap<>();

       // new ChildLine(10,20,100,100, Sex.FEMALE,new LinkedList<>());
        HashSet<Relationship> handledRelationships = new HashSet<>();
        for(LocalBlock localBlock : localBlocks){
            //todo: how to sort the persons inside a block?? right now they are sorted according to their birth date -> see Sorter.java
            for(Person person : localBlock){
                person.setDoiVisited(false);
                if(person.getCurrentDOI()<slider.getValue())
                    continue;


                double startX = timeAxis.getPixelForDate(person.getDate_of_birth());
                double endX = timeAxis.getPixelForDate(person.getDate_of_death());
                LifeLine lifeLine = new LifeLine(person, startX, currentHeight, endX, new ArrayList<>());

                lifeLinesOfPersons.put(person, lifeLine);
                initYPosition.put(person, currentHeight);
                currentYPosition.put(person, currentHeight);
                currentHeight += yDiff;

                spouseCountAbove.put(person, 0);
                spouseCountBelow.put(person, 0);
            }
        }


        List<Relationship> relationships = new ArrayList<>();
        for(LocalBlock localBlock : localBlocks) {
            //put all relationships of a block in one list and order them
            for (Person person : localBlock) {
                for (Relationship relationship : person.getRelationships()) {
                    if (!relationships.contains(relationship)) {
                        relationships.add(relationship);
                    }
                }
            }
        }
        relationships.sort((o1, o2) -> o1.getStart().before(o2.getStart()) ? -1 : o1.getStart().after(o2.getStart()) ? 1 : 0);

        //we are not looping over the persons in a block but over the relations in sorted order
        for(Relationship relationship : relationships){
            if(relationship.getPerson3() != null){
                if(relationship.getPerson1().getCurrentDOI()<slider.getValue() || relationship.getPerson2().getCurrentDOI()<slider.getValue()||relationship.getPerson3().getCurrentDOI()<slider.getValue())
                    continue;
            }else {
                if(relationship.getPerson1().getCurrentDOI()<slider.getValue() || relationship.getPerson2().getCurrentDOI()<slider.getValue())
                    continue;
            }

            if(handledRelationships.contains(relationship))
                continue;
            handledRelationships.add(relationship);

            Person person1 = relationship.getPerson1();
            Person person2 = relationship.getPerson2();
            Person person3 = relationship.getPerson3();
            LifeLine lifeLine1 = lifeLinesOfPersons.get(person1);
            LifeLine lifeLine2 = lifeLinesOfPersons.get(person2);
            Date date = relationship.getStart();
            double currentYPosition1 = currentYPosition.get(person1);
            double currentYPosition2 = currentYPosition.get(person2);

            if(relationship.getType() == RelationshipType.MARRIAGE){
                int marriages1 = currentMarriagesCounter(person1, date);
                int marriages2 = currentMarriagesCounter(person2, date);

                if(marriages1 > 1 && marriages2 > 1) {//cannot be visualized

                    continue;
                }else if(marriages1 > 1) {
                    //person2 has to go to person1
                    double yDiff2 = currentYPosition1 - currentYPosition2;

                    if(yDiff2 < 0) {//person2 goes upward
                        int marrySlots1 = spouseCountBelow.get(person1);
                        for(int i = 0; i < marrySlots1 + 1; i++)
                            yDiff2 += (LifeLine.STROKE_WIDTH + 3);
                        spouseCountBelow.replace(person1, marrySlots1, marrySlots1 + 1);
                        spouseCountAbove.replace(person2, 0, 1);
                    }else {//person2 goes downwards
                        int marrySlots1 = spouseCountAbove.get(person1);
                        for(int i = 0; i < marrySlots1 + 1; i++)
                            yDiff2 -= (LifeLine.STROKE_WIDTH + 3);
                        spouseCountAbove.replace(person1, marrySlots1, marrySlots1 + 1);
                        spouseCountBelow.replace(person2, 0, 1);
                    }

                    lifeLine2.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff2));
                    currentYPosition.replace(person2, currentYPosition2, currentYPosition2 + yDiff2);
                }else if(marriages2 > 1){
                    //person1 has to go to person2
                    double yDiff1 = currentYPosition2 - currentYPosition1;

                    if(yDiff1 < 0) {
                        int marrySlots2 = spouseCountBelow.get(person2);
                        for(int i = 0; i < marrySlots2 + 1; i++)
                            yDiff1 += (LifeLine.STROKE_WIDTH + 3);
                        spouseCountBelow.replace(person2, marrySlots2, marrySlots2 + 1);
                        spouseCountAbove.replace(person1, 0, 1);
                    }else {
                        int marrySlots2 = spouseCountAbove.get(person2);
                        for(int i = 0; i < marrySlots2 + 1; i++)
                            yDiff1 -= (LifeLine.STROKE_WIDTH + 3);
                        spouseCountAbove.replace(person2, marrySlots2, marrySlots2 + 1);
                        spouseCountBelow.replace(person1, 0, 1);
                    }
                    lifeLine1.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff1));
                    currentYPosition.replace(person1, currentYPosition1, currentYPosition1 + yDiff1);
                }else{
                    //both meet in the middle
                    double meetPosition = (currentYPosition1 + currentYPosition2)/2.0;//todo: find a way that meet position do not overlap with initial positions of persons...
                    double yDiff1 = meetPosition - currentYPosition1;
                    double yDiff2 = meetPosition - currentYPosition2;
                    if(yDiff1 < 0) {//person1 goes upwards
                        int marrySlots2 = spouseCountBelow.get(person2);
                        for(int i = 0; i < marrySlots2; i++)
                            yDiff1 += (LifeLine.STROKE_WIDTH + 3.0);
                        yDiff1 += (LifeLine.STROKE_WIDTH/2.0 + 1.5);
                    }else{
                        int marrySlots2 = spouseCountAbove.get(person2);
                        for(int i = 0; i < marrySlots2; i++)
                            yDiff1 -= (LifeLine.STROKE_WIDTH + 3.0);
                        yDiff1 -= (LifeLine.STROKE_WIDTH/2.0 + 1.5);
                    }
                    if(yDiff2 < 0) {
                        int marrySlots1 = spouseCountBelow.get(person1);
                        for(int i = 0; i < marrySlots1; i++)
                            yDiff2 += (LifeLine.STROKE_WIDTH + 3.0);
                        yDiff2 += (LifeLine.STROKE_WIDTH/2.0 + 1.5);
                    }else {
                        int marrySlots1 = spouseCountAbove.get(person1);
                        for(int i = 0; i < marrySlots1; i++)
                            yDiff2 -= (LifeLine.STROKE_WIDTH + 3.0);
                        yDiff2 -= (LifeLine.STROKE_WIDTH/2.0 + 1.5);
                    }

                    if(yDiff1 < 0){
                        int marrySlots2 = spouseCountAbove.get(person2);
                        spouseCountBelow.replace(person2, marrySlots2, marrySlots2 + 1);
                        spouseCountAbove.replace(person1, 0, 1);
                    }else{
                        int marrySlots2 = spouseCountAbove.get(person2);
                        spouseCountAbove.replace(person2, marrySlots2, marrySlots2 + 1);
                        spouseCountBelow.replace(person1, 0, 1);
                    }

                    lifeLine1.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff1));
                    lifeLine2.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff2));
                    currentYPosition.replace(person1, currentYPosition1, currentYPosition1 + yDiff1);
                    currentYPosition.replace(person2, currentYPosition2, currentYPosition2 + yDiff2);

                }

            }else if(relationship.getType() == RelationshipType.DIVORCE){

                double initYPosition1 = initYPosition.get(person1);
                double initYPosition2 = initYPosition.get(person2);
                double yDiff1 = initYPosition1 - currentYPosition1;
                double yDiff2 = initYPosition2 - currentYPosition2;
                lifeLine1.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff1));
                lifeLine2.addBend(new Bend(timeAxis.getPixelForDate(date), yDiff2));
                currentYPosition.replace(person1, currentYPosition1, initYPosition1);
                currentYPosition.replace(person2, currentYPosition2, initYPosition2);

            }else if(relationship.getType() == RelationshipType.BIRTH_OF_CHILD){
                List<Person> persons = new ArrayList<>();
                persons.add(person1);
                persons.add(person2);
                persons.add(person3);
                persons.sort((o1, o2) -> o1.getDate_of_birth().before(o2.getDate_of_birth()) ? 1 : o1.getDate_of_birth().after(o2.getDate_of_birth()) ? -1 : 0);
                Person child = persons.get(0);
                Person parent1 = persons.get(1);
                Person parent2 = persons.get(2);

                double childYPosition = currentYPosition.get(child);
                double parent1YPosition = currentYPosition.get(parent1);
                double parent2YPosition = currentYPosition.get(parent2);
                List<Double> circleHeights = new ArrayList<>();
                circleHeights.add(parent1YPosition);
                circleHeights.add(parent2YPosition);
                double startY = Math.min(Math.min(childYPosition, parent1YPosition), parent2YPosition);
                double endY = Math.max(Math.max(childYPosition, parent1YPosition), parent2YPosition);
                double x = timeAxis.getPixelForDate(child.getDate_of_birth());
                new ChildLine(x, startY, endY, childYPosition, child.getSex(), circleHeights);
            }
        }
        lifeLinesOfPersons.values().forEach(LifeLine::buildFinalLine);


        if(selectedLifeline == null){
            Person first = localBlocks.get(0).get(0);
            LifeLine  lifeLine = lifeLinesOfPersons.get(first);

            if(lifeLine != null){
                double x = lifeLine.getStartX();
                double y = lifeLine.getStartY();
                CustomBorderPane.DRAGGABLE_SHAPES.forEach(draggableShape1 -> {
                    draggableShape1.dragX(CustomBorderPane.instance().startDragX - x);
                    draggableShape1.dragY(CustomBorderPane.instance().startDragY - y);
                });
            }
        }else{
            double x = selectedLifeline.getStartX();
            double y = selectedLifeline.getStartY();
            CustomBorderPane.DRAGGABLE_SHAPES.forEach(draggableShape1 -> {
                draggableShape1.dragX(CustomBorderPane.instance().startDragX - x);
                draggableShape1.dragY(CustomBorderPane.instance().startDragY - y);
            });
        }
    }


    /**
     * Calculates the current amount of marriages of a person on a specific date.
     * @param person the person of which the marriages should be counted.
     * @param date the date at which the marriages should be counted.
     * @return the count of marriages of the person at the date.
     */
    private static int currentMarriagesCounter(Person person, Date date){

        List<Relationship> currentMarriages = new LinkedList<>();

        for(Relationship relationship : person.getRelationships()){
            if(relationship.getType() == RelationshipType.MARRIAGE){
                if(relationship.getStart().before(date) || relationship.getStart().equals(date))
                    currentMarriages.add(relationship);
            }

            if(relationship.getType() == RelationshipType.DIVORCE){
                if(relationship.getStart().before(date) || relationship.getStart().equals(date)){
                    for(Relationship currentMarriage : currentMarriages){
                        if(relationship.getPerson1() == currentMarriage.getPerson1() && relationship.getPerson2() == currentMarriage.getPerson2() ||
                                relationship.getPerson1() == currentMarriage.getPerson2() && relationship.getPerson2() == currentMarriage.getPerson1()){
                            currentMarriages.remove(currentMarriage);
                        }
                    }
                }
            }
        }
        return currentMarriages.size();
    }
}
