package main.java.datamodel;


import main.java.datamodel.exceptions.WrongFilePathException;
import org.gedcom4j.exception.GedcomParserException;
import org.gedcom4j.model.*;
import org.gedcom4j.model.enumerations.FamilyEventType;
import org.gedcom4j.model.enumerations.IndividualEventType;
import org.gedcom4j.parser.DateParser;
import org.gedcom4j.parser.GedcomParser;

import java.io.IOException;
import java.util.*;

/**
 * Object used to convert a Gedcom file for a given path
 */
public class GedcomConverter {

    /**
     * defines the path to the gedcom file
     */
    private String path;
    /**
     * a Gedcom4J class to parse a gedcom file
     */
    private GedcomParser gedcomParser;

    /**
     * Create a Gedcom Converter Object
     * @param path defines the absolute path to the gedcom file
     * @throws WrongFilePathException if the file was not found/ Is almost impossible as another method finds the absolute path in an input dialog
     */
    public GedcomConverter(String path) throws WrongFilePathException {

        this.path = path;
        gedcomParser = new GedcomParser();

        try {
            gedcomParser.load(path);
        } catch (IOException | GedcomParserException e) {
            throw new WrongFilePathException("File was not found");
        }
    }


    /**
     * iterates over the structure generated by Gedcom4j and converts the persons into our internal data format
     * additionaly the relationships are calculated from the family structure which is also generated by Gedcom4J
     * @return a Converterobject storing th persons and relationships
     */
    public ConverterObject parseGedcom(){

        Gedcom gedcom = gedcomParser.getGedcom();
        HashMap<Individual, Person> persons = new HashMap<>();
        int counter = 0;

        for (Individual i: gedcom.getIndividuals().values()) {
            Person person = new Person();
            person.setName(i.getNames().get(0).getBasic().replace("/",""));
            if (person.getName().toLowerCase().contains("unknown")){ //delete person if there is no name; in the royal92 gedcom file there are also person whose name is eg daughter !
                continue;
            }

            setSex(person, i);
            setBirthdate(person, i);
            setDeathDate(person,i);

            if(person.getDate_of_birth() != null){
                Calendar cal = Calendar.getInstance();
                cal.setTime(person.getDate_of_birth());
            }
            persons.put(i,person);
            counter++;
        }

        List<Relationship> relationships = new LinkedList<>();


        //extract marriages and divorces
        for (Family family : gedcom.getFamilies().values()) {
            if(family.getChildren() != null){
                for (IndividualReference i: family.getChildren()) {

                    if(family.getWife() == null){//todo now children where either the father or the mother is unknown are not saved correctly (the relationship is not saved at all)
                        continue;
                    }
                    if(family.getHusband() == null){
                        continue;
                    }
                    Individual wifeName = family.getWife().getIndividual();
                    Individual husbandName = family.getHusband().getIndividual();
                    Individual childName = i.getIndividual();

                    //if mother or child is not saved do not draw the relationship
                    if(persons.containsKey(wifeName) && persons.containsKey(childName) && persons.containsKey(husbandName)){
                        Person mother = persons.get(wifeName);
                        Person child = persons.get(childName);
                        Person father = persons.get(husbandName);

                        Relationship relationship = new Relationship(mother,child,father);

                        relationship.setType(RelationshipType.BIRTH_OF_CHILD);
                        relationship.setApproximateDate(child.isBirthDateApproximated());
                        relationship.setStart(child.getDate_of_birth());
                        relationships.add(relationship);
                        checkForFamilyApproximations(family,child,persons);

                    }
                }
            }

            if(family.getEvents() == null){
                continue;
            }else {
                for (FamilyEvent event : family.getEvents()) {
                    // find husband
                    if(family.getHusband() == null){
                        continue;
                    }

                    Individual husbandName = family.getHusband().getIndividual();//.getNames().get(0).getBasic().replace("/","");
                    if(persons.containsKey(husbandName)){// check if this person is already registered

                        // find wife if she is not there jump over this loop (= do not save this relationship)
                        if(family.getWife() == null){

                            continue;
                        }

                        Individual wifeName = family.getWife().getIndividual();//;.getNames().get(0).getBasic().replace("/","");
                        if(persons.containsKey(wifeName)){// check if this person is already registered
                                //now we have found the husband and the wife; we can also find the children of this family
                            Person person1 = persons.get(husbandName);
                            Person person2 = persons.get(wifeName);
                            Relationship relationship = new Relationship(person1,person2);

                            if(event.getType() == FamilyEventType.MARRIAGE){

                                if(person1.getDate_of_birth() == null && person2.getDate_of_birth() != null){
                                    person1.setBirthDateApproximated(true);
                                    person1.setDate_of_birth(person2.getDate_of_birth());
                                }
                                if(person2.getDate_of_birth() == null && person1.getDate_of_birth() != null){
                                    person2.setBirthDateApproximated(true);
                                    person2.setDate_of_birth(person1.getDate_of_birth());
                                }

                                relationship.setType(RelationshipType.MARRIAGE);

                            }else if(event.getType() == FamilyEventType.DIVORCE){
                                relationship.setType(RelationshipType.DIVORCE);
                            }

                            setRelationshipDate(relationship,event);
                            relationships.add(relationship);
                        }
                    }

                }
            }
        }
        return new ConverterObject(persons,relationships);
    }

    /**
     * approximates missing values directly from the family structure
     * the birth/ death date of a person with the spouses birth/ death date
     * a childs birth/death date with the siblings mean birth date
     * @param family
     * @param child
     * @param persons
     */
    private void checkForFamilyApproximations(Family family, Person child, HashMap<Individual, Person> persons) {

        //approximate birthdate with marriage date of the paretns
        if(child.getDate_of_birth() == null){
            if(family.getEvents() == null){
                return;
            }
            for (FamilyEvent event : family.getEvents()) {
                if(event.getType() == FamilyEventType.MARRIAGE){

                    Person wife = null;
                    Person husband = null;

                    // approximate spouses birth with birthdate of husband or wife
                    if(family.getHusband()!=null){
                        //String husbandname = family.getHusband().getIndividual().getNames().get(0).getBasic().replace("/","");
                        husband = persons.get(family.getHusband());
                    }
                    if(family.getWife()!= null){
                        //String wifename = family.getWife().getIndividual().getNames().get(0).getBasic().replace("/","");
                        wife = persons.get(family.getWife());
                    }

                    if(family.getHusband()==null && family.getWife() == null || wife == null && husband == null){
                        continue;
                    }

                    if(wife.getDate_of_birth() == null &&
                        husband.getDate_of_birth() != null){
                        wife.setBirthDateApproximated(true);
                        wife.setDate_of_birth(husband.getDate_of_birth());
                    }
                    if(husband.getDate_of_birth() == null && wife.getDate_of_birth() != null){
                        husband.setBirthDateApproximated(true);
                        husband.setDate_of_birth(wife.getDate_of_birth());
                    }

                    //we dont need the persons here this object is only used to get the date
                    Relationship relationship = new Relationship();
                    setRelationshipDate(relationship,event);
                    child.setBirthDateApproximated(true);
                    child.setDate_of_birth(relationship.getStart());
                }
            }
        }

        //approximate death date by mean death of the siblings
        if(child.getDate_of_death() == null){

            //get other children of family
            for (IndividualReference i: family.getChildren()) {
                Individual currentIndividual = i.getIndividual();
                String currentName = i.getIndividual().getNames().get(0).getBasic().replace("/","");
                long sum = 0;
                int counter = 0;
                //only add the siblings
                if(!currentName.equals(child.getName())){//todo should we compare the names or the gedcom4j.Individuals
                    Person sibling = persons.get(currentIndividual);
                    if(sibling != null){
                        if(sibling.getDate_of_death() != null){
                            sum += sibling.getDate_of_death().getTime();
                            counter ++;
                        }
                    }
                }
                child.setDeathDateApproximated(true);

                if(counter>0){// avoid division by zero exception todo think about another approximation
                    child.setDate_of_death(new Date(sum/counter));
                }
            }
        }
    }

    /**
     * this method is used to set the sex of a person
     * @param person the person
     * @param i the indidviudal, which is a representation of a person from Gedcom4J
     */
    private void setSex(Person person, Individual i){

        if(i.getSex() != null){
            if(i.getSex().getValue().toLowerCase().equals("m")){
                person.setSex(Sex.MALE);
            }else if (i.getSex().getValue().toLowerCase().equals("f")){
                person.setSex(Sex.FEMALE);
            }
        }
    }
    /**
     * this method is used to set birth date of a person
     * @param person the person
     * @param i the indidviudal, which is a representation of a person from Gedcom4J
     */
    private void setBirthdate(Person person,Individual i){

        int size = i.getEventsOfType(IndividualEventType.BIRTH).size();
        if(size == 0){
            person.setBirthDateApproximated(true);
            return;
        }

        for (IndividualEvent e : i.getEventsOfType(IndividualEventType.BIRTH)) {
            if(e.getDate()!= null){
                person.setDate_of_birth(new DateParser().parse(e.getDate().toString()));
            }else {// the date is null set
                person.setBirthDateApproximated(true);
            }
        }
    }
    /**
     * this method is used to set death date of a person
     * @param person the person
     * @param i the indidviudal, which is a representation of a person from Gedcom4J
     */
    private void setDeathDate(Person person,Individual i){

        int size = i.getEventsOfType(IndividualEventType.DEATH).size();
        if(size == 0){
            person.setDeathDateApproximated(true);
        }

        for (IndividualEvent e : i.getEventsOfType(IndividualEventType.DEATH)) {
            if(e.getDate()!= null){
                person.setDate_of_death(new DateParser().parse(e.getDate().toString()));
            }else {// the date is null set
                person.setDeathDateApproximated(true);
            }
        }
    }

    /**
     * this method is used to set the date of a relationship
     * @param relationship the relationship with the date still missing
     * @param event the relationship/ event representation from the Gedcom4J
     */
    private void setRelationshipDate(Relationship relationship,FamilyEvent event){

        if(event.getDate()!= null){
            relationship.setStart(new DateParser().parse(event.getDate().toString()));
        }else {// the date is null set
            relationship.setApproximateDate(true);
        }
    }
}
