package main.java.layout;

import main.java.datamodel.ConverterObject;
import main.java.datamodel.Person;
import main.java.datamodel.Relationship;
import main.java.datamodel.RelationshipType;
import org.gedcom4j.model.Individual;

import java.util.*;

/**
 * This class contains the necessary functionality to approximate the different dates, including birth and death of persons as well as marriage and divorce of spouses.
 */
public class Approximation {

    /**
     * Approximates the missing data in the person and relationship objects, which converterObject contains.
     * @param converterObject the converterObject containing the persons and the relationships.
     */
    public static void approximate(ConverterObject converterObject){

        for(Person person : converterObject.getPersons().values())
            approximate(person);

        for(Relationship relationship : converterObject.getRelationship())
            approximate(relationship);
    }

    /**
     * Approximates the missing data of the specific person.
     * @param person the person.
     */
    private static void approximate(Person person){

        approximateBirth(person);
        approximateDeath(person);
    }

    /**
     * approximates the missing data of the specific relationship
     * @param relationship the relationship
     */
    private static void approximate(Relationship relationship){

        if(relationship.getType() == RelationshipType.MARRIAGE)
            approximateMarriage(relationship);
        if(relationship.getType() == RelationshipType.BIRTH_OF_CHILD)
            approximateChildBirth(relationship);
        if(relationship.getType() == RelationshipType.DIVORCE)
            approximateDivorce(relationship);
    }

    /**
     * approximates the birthdate of the person by looking at different relationships of the person
     * @param person the person
     */
    private static void approximateBirth(Person person){

        if(person.getDate_of_birth() != null)
            return;

        person.setBirthDateApproximated(true);

        for(Relationship relationship : person.getRelationships()) {

            if (relationship.getType() == RelationshipType.MARRIAGE || relationship.getType() == RelationshipType.DIVORCE) {
                Person spouse = person == relationship.getPerson1() ? relationship.getPerson2() : relationship.getPerson1();

                Date spouseBirth = spouse.getDate_of_birth();
                Date personBirth = person.getDate_of_birth();
                Date relationDate = relationship.getStart();


                if (spouseBirth != null) {
                    if (personBirth == null || personBirth.after(spouseBirth)) {
                        person.setDate_of_birth(spouseBirth);                                                               //approximate the birth date with the oldest spouse birth date
                    }
                } else if (relationDate != null) {
                    Calendar calendar = Calendar.getInstance();
                    calendar.setTime(relationDate);
                    calendar.set(Calendar.YEAR, calendar.get(Calendar.YEAR) - 20);                                          //or the earliest relationship date minus 20 years.
                    if (personBirth == null || personBirth.after(calendar.getTime())) {
                        person.setDate_of_birth(calendar.getTime());

                    }
                }
            }
        }

        if(person.getDate_of_birth() == null && person.getDate_of_death() != null){
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(person.getDate_of_death());
            calendar.set(Calendar.YEAR, calendar.get(Calendar.YEAR) - 80);                                              //or with its death date
            person.setDate_of_birth(calendar.getTime());

        }
    }


    /**
     * approximates the deathdate of the person by looking at different relationships of the person
     * @param person the person
     */
    private static void approximateDeath(Person person){

        if(person.getDate_of_death() != null)
            return;

        person.setDeathDateApproximated(true);

        for(Relationship relationship : person.getRelationships()){

            if (relationship.getType() == RelationshipType.MARRIAGE || relationship.getType() == RelationshipType.DIVORCE) {
                Person spouse = person == relationship.getPerson1() ? relationship.getPerson2() : relationship.getPerson1();//in all relationships this will be the spouse of the person

                Date spouseDeath = spouse.getDate_of_death();
                Date personDeath = person.getDate_of_death();
                Date relationDate = relationship.getStart();

                if (spouseDeath != null) {
                    if (personDeath == null || personDeath.before(spouseDeath)) {
                        person.setDate_of_death(spouseDeath);                                                               //approximate the death date with the spouse who lives the longest
                    }
                } else if (relationDate != null) {
                    Calendar tCalendar = Calendar.getInstance();
                    tCalendar.setTime(relationDate);
                    tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) + 5);                                         //or the latest relationship  date plus 5 years.
                    if (personDeath == null || personDeath.after(tCalendar.getTime())) {
                        person.setDate_of_death(tCalendar.getTime());
                    }
                }
            }
        }

        if(person.getDate_of_death() == null && person.getDate_of_birth() != null){
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(person.getDate_of_birth());
            calendar.set(Calendar.YEAR, calendar.get(Calendar.YEAR) + 80);                                              //or with its birth date
            person.setDate_of_death(calendar.getTime());

        }
    }


    /**
     * Approximates the date of the marriage by looking at the two spouses.
     * @param marriage the marriage relationship
     */
    private static void approximateMarriage(Relationship marriage){

        if(marriage.getStart() != null)
            return;

        Person person1 = marriage.getPerson1();
        Person person2 = marriage.getPerson2();

        for(Relationship childBirth : person1.getRelationships()){
            if(childBirth.getType() == RelationshipType.BIRTH_OF_CHILD &&
                    (childBirth.getPerson1() == person1 && childBirth.getPerson2() == person2 ||
                            childBirth.getPerson1() == person2 && childBirth.getPerson2() == person1)){
                Date childBirthDate = childBirth.getStart();
                if(childBirthDate != null){
                    Calendar tCalendar = Calendar.getInstance();
                    tCalendar.setTime(childBirthDate);
                    tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) - 1);                                     //approximate the marriage by the childbirth -1 year
                    marriage.setStart(tCalendar.getTime());

                    break;
                }
            }
        }

        if(marriage.getStart() != null)
            return;

        for(Relationship divorce : person1.getRelationships()){
            if(divorce.getType() == RelationshipType.DIVORCE &&
                    (divorce.getPerson1() == person1 && divorce.getPerson2() == person2 ||
                            divorce.getPerson1() == person2 && divorce.getPerson2() == person1)){
                Date divorceDate = divorce.getStart();
                if(divorceDate != null){
                    Calendar tCalendar = Calendar.getInstance();
                    tCalendar.setTime(divorceDate);
                    tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) - 2);                                     //or approximate the marriage by assume that the marriage lasts for 2 years.
                    marriage.setStart(tCalendar.getTime());
                    break;
                }
            }
        }


        if(marriage.getStart() != null)
            return;

        Date person1Birth = person1.getDate_of_birth();
        Date person2Birth = person2.getDate_of_birth();
        if(person1Birth != null && person2Birth != null) {
            Date youngerBirth = person1Birth.after(person2Birth) ? person1Birth : person2Birth;
            Calendar tCalendar = Calendar.getInstance();
            tCalendar.setTime(youngerBirth);
            tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) + 18);                                            //or approximate by the assumption that the younger person was 18.
            marriage.setStart(tCalendar.getTime());
        }else if(person1Birth != null){
            Calendar tCalendar = Calendar.getInstance();
            tCalendar.setTime(person1Birth);
            tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) + 18);                                            //or the given one was 18.
            marriage.setStart(tCalendar.getTime());
        }else if(person2Birth != null){
            Calendar tCalendar = Calendar.getInstance();
            tCalendar.setTime(person2Birth);
            tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) + 18);                                            //same as above.
            marriage.setStart(tCalendar.getTime());
        }

    }


    /**
     * Makes sure that the birthdate of the child and the relationship date are the same.
     * @param childbirth the childBirth relationship
     */
    private static void approximateChildBirth(Relationship childbirth){

        Person person1 = childbirth.getPerson1();
        Person person2 = childbirth.getPerson2();
        Person person3 = childbirth.getPerson3();

        Date birthDate1 = person1.getDate_of_birth();
        Date birthDate2 = person2.getDate_of_birth();
        Date birthDate3 = person3.getDate_of_birth();

        if(birthDate1 != null && birthDate2 != null && birthDate3 != null){

            List<Date> birthDates = new ArrayList<>();
            birthDates.add(birthDate1);
            birthDates.add(birthDate2);
            birthDates.add(birthDate3);
            birthDates.sort((o1, o2) -> o1.before(o2) ? -1 : o1.after(o2) ? 1 : 0);

            Date childBirthDate = birthDates.get(2);
            childbirth.setStart(childBirthDate);
        }else{
            childbirth.setStart(null);
        }
    }


    /**
     * Approximates the date of the divorce by looking at the two spouses.
     * @param divorce the divorce relationship
     */
    private static void approximateDivorce(Relationship divorce){

        if(divorce.getStart() != null)
            return;

        Person person1 = divorce.getPerson1();
        Person person2 = divorce.getPerson2();

        for(Relationship marriage : person1.getRelationships()){
            if(marriage.getType() == RelationshipType.MARRIAGE &&
                    (marriage.getPerson1() == person1 && marriage.getPerson2() == person2 ||
                            marriage.getPerson1() == person2 && marriage.getPerson2() == person1)){
                Date marriageDate = marriage.getStart();
                if(marriageDate != null){
                    Calendar tCalendar = Calendar.getInstance();
                    tCalendar.setTime(marriageDate);
                    tCalendar.set(Calendar.YEAR, tCalendar.get(Calendar.YEAR) + 2);                                     //approximate divorce by marriage date
                    divorce.setStart(tCalendar.getTime());
                    break;
                }
            }
        }
    }
}
