package main.java.layout;

import main.java.datamodel.ConverterObject;
import main.java.datamodel.Person;
import main.java.datamodel.Relationship;
import org.gedcom4j.model.Individual;

import java.util.*;

/**
 * This class contains the necessary functionality to sort out all objects with incomplete data and sort the remaining one according to the date.
 */
public class Sorter {

    /**
     * This method removes all the objects with incomplete data, i.e. which were not approximated correctly.
     * @param converterObject the converterObject containing all persons and relationships.
     */
    public static void sortOutAllFailedApproximations(ConverterObject converterObject){

        Set<Map.Entry<Individual, Person>> personsEntriesToDelete = new HashSet<>();
        Set<Person> personsToDelete = new HashSet<>();
        Set<Relationship> relationshipsToDelete = new HashSet<>();

        for(Map.Entry<Individual, Person> entry : converterObject.getPersons().entrySet()){
            Person person = entry.getValue();
            if(person.getDate_of_birth() == null || person.getDate_of_death() == null || person.getName() == null || person.getSex() == null) {
                personsEntriesToDelete.add(entry);
                personsToDelete.add(person);
            }
        }

        for(Relationship relationship : converterObject.getRelationship()){
            if(relationship.getStart() == null || relationship.getPerson1() == null || relationship.getPerson2() == null || relationship.getType() == null)
                relationshipsToDelete.add(relationship);
        }

        for(Relationship relationship : converterObject.getRelationship()){
            if(personsToDelete.contains(relationship.getPerson1()) || personsToDelete.contains(relationship.getPerson2())){
                relationshipsToDelete.add(relationship);
            }
        }

        for(Person person : converterObject.getPersons().values()){
            for(Relationship relationship : relationshipsToDelete){
                if(person.getRelationships().contains(relationship)){
                    person.getRelationships().remove(relationship);
                }
            }
        }

        for(Map.Entry<Individual, Person> entry : personsEntriesToDelete){
            converterObject.getPersons().remove(entry.getKey(), entry.getValue());
        }

        for(Relationship relationship : relationshipsToDelete){
            converterObject.getRelationship().remove(relationship);
        }
    }


    /**
     * Sorts the persons inside the local blocks according to their birth date.
     * @param localBlocks a list of all local blocks.
     */
    public static void sortBlocks(List<LocalBlock> localBlocks){

        for(LocalBlock localBlock : localBlocks){
            localBlock.sort((o1, o2) -> o1.getDate_of_birth().before(o2.getDate_of_birth()) ? -1 : o1.getDate_of_birth().after(o2.getDate_of_birth()) ? 1 : 0);
        }
    }

    /**
     * Sorts the relationships according to their date.
     * @param relationships a list containing all relationships.
     */
    public static void sortRelationships(List<Relationship> relationships){

        relationships.sort((o1, o2) -> o1.getStart().before(o2.getStart()) ? -1 : o1.getStart().after(o2.getStart()) ? 1 : 0);
    }

    /**
     * Sorts the relationships of a person according to their date.
     * @param persons the person containing the relationships.
     */
    public static void sortRelationshipsInPersons(Map<Individual, Person> persons){

        for(Person person : persons.values()){
            person.getRelationships().sort((o1, o2) -> o1.getStart().before(o2.getStart()) ? -1 : o1.getStart().after(o2.getStart()) ? 1 : 0);
        }
    }
}