package main.java.shape;

import javafx.scene.paint.Color;
import javafx.scene.paint.CycleMethod;
import javafx.scene.paint.LinearGradient;
import javafx.scene.paint.Stop;
import javafx.scene.shape.*;
import javafx.scene.text.Text;
import main.java.CustomBorderPane;
import main.java.Main;
import main.java.datamodel.Person;
import main.java.datamodel.Sex;
import main.java.shape.abs.DraggableShape;
import main.java.shape.abs.SelectableNode;
import main.java.utils.Colors;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

/**
 * This class represents a whole life line of a person containing all bends and other necessary data.
 */
public class LifeLine extends Path implements SelectableNode, DraggableShape {

    /**
     * The width of the life lines.
     */
    public static final int STROKE_WIDTH = 10;


    private Person person;
    private boolean selected = false;
    private Text text;
    private String name;
    private List<MoveTo> moveTos = new ArrayList<>();
    private List<LineTo> lineTos = new ArrayList<>();
    private List<CubicCurveTo> cubicCurveTos = new ArrayList<>();
    private List<Bend> bends = new ArrayList<>();
    private double startX;
    private double startY;
    private double endX;

    /**
     * The contructor to create a life line
     * @param person the person which the life line should represent.
     * @param startX the initial x coordinate of the life line in pixels.
     * @param startY the initial y coordinate of the life line in pixels.
     * @param endX the final x coordinate of the life line in pixels.
     * @param bends a list of bends the lifeline should contain. Has to be sorted according to their x coordinate.
     */
    public LifeLine(Person person, double startX, double startY, double endX, List<Bend> bends){

        super();
        this.person = person;
        this.name = person.getName();
        text = new Text(name);
        text.setX(startX);
        text.setY(startY - STROKE_WIDTH/2.0 - 2);
        this.startX = startX;
        this.startY = startY;
        this.endX = endX;
        this.bends = bends;

        this.setStrokeWidth(STROKE_WIDTH);
        CustomBorderPane.DRAGGABLE_SHAPES.add(this);
        
        CustomBorderPane.instance().getChildren().add(this);
        CustomBorderPane.instance().getChildren().add(text);

        setColor(false);
    }


    public Person getPerson() {
        return person;
    }

    public void setPerson(Person person) {
        this.person = person;
    }

    /**
     * Sets the color of the life line according to the sex of the person changes it if the life line is selected.
     * @param selected
     */
    private void setColor(boolean selected){

        Color colorM = Colors.COLOR_M;
        Color colorF = Colors.COLOR_F;
        Color colorMT = Colors.COLOR_M_TRANSPARENT;
        Color colorFT = Colors.COLOR_F_TRANSPARENT;


        if(selected){
            colorM = Colors.COLOR_M.brighter();
            colorF = Colors.COLOR_F.brighter();
            colorMT = Colors.COLOR_M_TRANSPARENT.brighter();
            colorFT = Colors.COLOR_F_TRANSPARENT.brighter();
        }

        List<Stop> stops = new LinkedList<>();

        if(person.isBirthDateApproximated() && person.getSex() == Sex.MALE){
            stops.add(new Stop(0.0, colorMT));
            stops.add(new Stop(0.05, colorM));
        }
        if(person.isBirthDateApproximated() && person.getSex() == Sex.FEMALE){
            stops.add(new Stop(0.0, colorFT));
            stops.add(new Stop(0.05, colorF));
        }
        if(person.isDeathDateApproximated() && person.getSex() == Sex.MALE){
            stops.add(new Stop(0.95, colorM));
            stops.add(new Stop(1, colorMT));
        }
        if(person.isDeathDateApproximated() && person.getSex() == Sex.FEMALE){
            stops.add(new Stop(0.95, colorF));
            stops.add(new Stop(1, colorFT));
        }
        if(!stops.isEmpty()) {
            this.setStroke(new LinearGradient(0, 0, 1, 0, true, CycleMethod.NO_CYCLE, stops));
        }else{
            if(person.getSex() == Sex.MALE){
                this.setStroke(new LinearGradient(0, 0, 1, 0, true, CycleMethod.NO_CYCLE, new Stop(0, colorM), new Stop(1, colorM)));
            }else{
                this.setStroke(new LinearGradient(0, 0, 1, 0, true, CycleMethod.NO_CYCLE, new Stop(0, colorF), new Stop(1, colorF)));
            }
        }
    }

    /**
     * Builds the final lifeline and adds it to the border pane such that it is drawn.
     */
    public void buildFinalLine(){
        this.getElements().add(new MoveTo(startX, startY));
        double currentY = startY;
        bends.sort(Comparator.comparingDouble(Bend::getStartX));
        for(Bend bend : bends){
            this.getElements().add(new LineTo(bend.getStartX(), currentY));
            this.createBend(bend.getStartX()+(Bend.BEND_WIDTH/3.0f),
                    currentY,
                    bend.getStartX()+2*(Bend.BEND_WIDTH/3.0f),
                    currentY+bend.getYDifference(),
                    bend.getStartX()+Bend.BEND_WIDTH,
                    currentY+bend.getYDifference());

            currentY = currentY + bend.getYDifference();
        }

        this.getElements().add(new LineTo(endX, currentY));

        for(PathElement pathElement : this.getElements()){
            if(pathElement instanceof MoveTo)
                moveTos.add((MoveTo)pathElement);
            if(pathElement instanceof LineTo)
                lineTos.add((LineTo)pathElement);
            if(pathElement instanceof CubicCurveTo)
                cubicCurveTos.add((CubicCurveTo)pathElement);
        }
    }

    /**
     * Creates a bend according to the parameters and adds it to the life line
     * @param controlX1 x coordinate of the control point 1
     * @param controlY1 y coordinate of the control point 1
     * @param controlX2 x coordinate of the control point 2
     * @param controlY2 y coordinate of the control point 2
     * @param endX the x coordinate of the end of the bend
     * @param endY the y coordinate of the end of the bend
     */
    private void createBend(double controlX1, double controlY1, double controlX2, double controlY2, double endX, double endY){

        CubicCurveTo cubicCurveTo = new CubicCurveTo();
        cubicCurveTo.setControlX1(controlX1);
        cubicCurveTo.setControlY1(controlY1);
        cubicCurveTo.setControlX2(controlX2);
        cubicCurveTo.setControlY2(controlY2);
        cubicCurveTo.setX(endX);
        cubicCurveTo.setY(endY);

        this.getElements().addAll(cubicCurveTo);
    }



    @Override
    public boolean requestSelection(boolean select) {
        return true;
    }

    @Override
    public void notifySelection(boolean select) {

        if(select && !selected){
            Main.decreaseDOIRoot(person);
            setColor(select);
        }
        if(!select && selected){
            Main.selectedLifeline = null;
            setColor(select);
        }

        selected = select;
    }


    @Override
    public void dragX(double xDiff){

        this.startX += xDiff;

        for(MoveTo moveTo : moveTos){
            double x = moveTo.getX();
            moveTo.setX(x + xDiff);
        }
        for(LineTo lineTo : lineTos){
            double x = lineTo.getX();
            lineTo.setX(x + xDiff);
        }
        for(CubicCurveTo cubicCurveTo : cubicCurveTos){
            double controlX1 = cubicCurveTo.getControlX1();
            double controlX2 = cubicCurveTo.getControlX2();
            double x = cubicCurveTo.getX();
            cubicCurveTo.setControlX1(controlX1 + xDiff);
            cubicCurveTo.setControlX2(controlX2 + xDiff);
            cubicCurveTo.setX(x + xDiff);
        }

        text.setX(text.getX() + xDiff);
    }


    @Override
    public void dragY(double yDiff){

        this.startY += yDiff;

        for(MoveTo moveTo : moveTos){
            double y = moveTo.getY();
            moveTo.setY(y + yDiff);
        }
        for(LineTo lineTo : lineTos){
            double y = lineTo.getY();
            lineTo.setY(y + yDiff);
        }
        for(CubicCurveTo cubicCurveTo : cubicCurveTos){
            double controlY1 = cubicCurveTo.getControlY1();
            double controlY2 = cubicCurveTo.getControlY2();
            double y = cubicCurveTo.getY();
            cubicCurveTo.setControlY1(controlY1 + yDiff);
            cubicCurveTo.setControlY2(controlY2 + yDiff);
            cubicCurveTo.setY(y + yDiff);
        }

        text.setY(text.getY() + yDiff);
    }


    public String getName() {
        return name;
    }

    /**
     * adds a bend to the lifeline. Has to be called before buildFinalLine()
     * @param bend the bend that should be added.
     */
    public void addBend(Bend bend){
        this.bends.add(bend);
    }

    public double getStartX(){
        return startX;
    }

    public double getStartY() {
        return startY;
    }
}
