package main.java.shape;

import javafx.beans.property.ReadOnlyDoubleProperty;
import javafx.scene.paint.Color;
import javafx.scene.shape.Line;
import javafx.scene.text.Text;
import main.java.CustomBorderPane;
import main.java.shape.abs.DraggableShape;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

/**
 * This class represents and constructs the time axis.
 */
public class TimeAxis implements DraggableShape {
    /**
     * The value of a year in pixels.
     */
    public static final int YEAR_IN_PIXELS = 20;
    private static final int STROKE_WIDTH = 2;
    private static final Color STROKE = Color.color(0.5, 0.5, 0.5);
    private static final int HEIGHT = 700;
    private static final int DECADE_MARKER_HEIGHT = 5;

    private List<Line> lines = new ArrayList<>();
    private Line axis;
    private List<Text> texts = new ArrayList<>();

    private int startYear;
    private int endYear;

    /**
     * Contructor to build a time axis and register it to the border pane, such that it is rendered.
     * @param startYear the start year of the time axis.
     * @param endYear the end year of the time axis.
     */
    public TimeAxis(int startYear, int endYear){

        this.startYear = startYear;
        this.endYear = endYear;

        int yearDiff = endYear - startYear;
        axis = new Line(0, HEIGHT, yearDiff*YEAR_IN_PIXELS, HEIGHT);
        axis.setStrokeWidth(STROKE_WIDTH);
        axis.setStroke(STROKE);

        for(int year = startYear, yearCounter = 0; year <= endYear; year++, yearCounter++){
            if(year%10 == 0){
                Line decadeMarker = new Line(yearCounter*YEAR_IN_PIXELS, HEIGHT, yearCounter*YEAR_IN_PIXELS, HEIGHT + DECADE_MARKER_HEIGHT);
                decadeMarker.setStrokeWidth(STROKE_WIDTH);
                decadeMarker.setStroke(STROKE);
                lines.add(decadeMarker);

                Text decadeText = new Text(String.format("%4d", year));
                decadeText.setStroke(STROKE);
                decadeText.setFill(STROKE);
                decadeText.setX(yearCounter*YEAR_IN_PIXELS - 13);
                decadeText.setY(HEIGHT + 20);
                texts.add(decadeText);
            }
        }

        CustomBorderPane.DRAGGABLE_SHAPES.add(this);
        CustomBorderPane.instance().getChildren().add(axis);
        CustomBorderPane.instance().getChildren().addAll(lines);
        CustomBorderPane.instance().getChildren().addAll(texts);
    }


    @Override
    public void dragX(double xDiff) {

        double startX = axis.getStartX();
        double endX = axis.getEndX();
        axis.setStartX(startX + xDiff);
        axis.setEndX(endX + xDiff);

        for(Line line : lines){
            startX = line.getStartX();
            endX = line.getEndX();
            line.setStartX(startX + xDiff);
            line.setEndX(endX + xDiff);
        }
        for(Text text : texts){
            double x = text.getX();
            text.setX(x + xDiff);
        }
    }


    @Override
    public void dragY(double yDiff) {
        //should not be draggable
    }

    /**
     * Sets the height property such that the y coordinate is always adjusted to the height properties value.
     * @param heightProperty
     */
    public void setHeightProperty(ReadOnlyDoubleProperty heightProperty){

        axis.startYProperty().bind(heightProperty.subtract(40));
        axis.endYProperty().bind(heightProperty.subtract(40));

        for(Line line : lines){
            line.startYProperty().bind(heightProperty.subtract(40));
            line.endYProperty().bind(heightProperty.subtract(30));

        }
        for(Text text : texts){
            text.yProperty().bind(heightProperty.subtract(15));
        }
    }

    /**
     * Returns the x coordinate in pixels for a given date.
     * @param date the date that should be converted.
     * @return the x coordinate in pixels for the specific date.
     */
    public double getPixelForDate(Date date){

        double pixelsPerYear = YEAR_IN_PIXELS;
        double pixelsPerMonth = YEAR_IN_PIXELS/12.0;
        double pixelsPerDay = pixelsPerMonth/30.0;

        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        int year = calendar.get(Calendar.YEAR);
        int month = calendar.get(Calendar.MONTH);
        int day = calendar.get(Calendar.DAY_OF_MONTH);

        year -= startYear;
        double sum = 0;
        sum += year*pixelsPerYear;
        sum += month*pixelsPerMonth;
        sum += day*pixelsPerDay;

        return sum;
    }
}
